import typing
from itertools import chain
from threading import Lock, RLock

import attr

from applitools.common import (
    EyesError,
    RenderInfo,
    RenderRequest,
    RenderStatus,
    RGridDom,
    VGResource,
    VisualGridSelector,
    logger,
)
from applitools.common.utils import datetime_utils, urljoin, urlparse
from applitools.selenium import css_parser

from .vg_task import VGTask

if typing.TYPE_CHECKING:
    from typing import Callable, Dict, Any, Text, List, Optional
    from applitools.common import RenderStatusResults, Region
    from applitools.selenium.visual_grid import RunningTest, EyesConnector


@attr.s(hash=True)
class RenderTask(VGTask):
    MAX_FAILS_COUNT = 5
    MAX_ITERATIONS = 100

    script = attr.ib(hash=False, repr=False)  # type: Dict[str, Any]
    running_test = attr.ib()  # type: RunningTest
    resource_cache = attr.ib(hash=False, repr=False)
    put_cache = attr.ib(hash=False, repr=False)
    eyes_connector = attr.ib(hash=False, repr=False)  # type: EyesConnector
    rendering_info = attr.ib()
    region_selectors = attr.ib(
        hash=False, factory=list
    )  # type: List[List[VisualGridSelector]]
    size_mode = attr.ib(default=None)
    region_to_check = attr.ib(hash=False, default=None)  # type: Region
    script_hooks = attr.ib(hash=False, default=None)  # type: Optional[Dict]
    agent_id = attr.ib(default=None)  # type: Optional[Text]
    selector = attr.ib(hash=False, default=None)  # type: Optional[Text]
    func_to_run = attr.ib(default=None, hash=False, repr=False)  # type: Callable

    def __attrs_post_init__(self):
        # type: () -> None
        self.func_to_run = self.perform  # type: Callable
        self.all_blobs = []  # type: List[VGResource]
        self.request_resources = {}  # type: Dict[Text, VGResource]
        self.resource_urls = []  # type: List[Text]
        self.discovered_resources_lock = RLock()

    def perform(self):
        # type: () -> RenderStatusResults

        def get_and_put_resource(url):
            # type: (str) -> VGResource
            resource = self.request_resources.get(url)
            self.eyes_connector.render_put_resource(running_render, resource)
            return resource

        requests = []
        rq = self.prepare_data_for_rg(self.script)
        requests.append(rq)
        fetch_fails = 0
        render_requests = None
        while True:

            try:
                render_requests = self.eyes_connector.render(*requests)
            except Exception as e:
                logger.exception(e)
                fetch_fails += 1
                datetime_utils.sleep(1500)
            if not render_requests:
                continue

            running_render = render_requests[0]
            rq.render_id = running_render.render_id
            need_more_dom = running_render.need_more_dom
            need_more_resources = (
                running_render.render_status == RenderStatus.NEED_MORE_RESOURCE
            )
            still_running = (
                need_more_resources
                or need_more_dom
                or fetch_fails > self.MAX_FAILS_COUNT
            )

            dom_resource = rq.dom.resource

            if need_more_resources:
                for url in running_render.need_more_resources:
                    self.put_cache.fetch_and_store(url, get_and_put_resource)

            if need_more_dom:
                self.eyes_connector.render_put_resource(running_render, dom_resource)

            if not still_running:
                break

        statuses = self.poll_render_status(rq)
        if statuses and statuses[0].status == RenderStatus.ERROR:
            raise EyesError(
                "Render failed for {} with the message: {}".format(
                    statuses[0].status, statuses[0].error
                )
            )
        return statuses[0]

    def prepare_data_for_rg(self, data):
        # type: (Dict) -> RenderRequest
        self.request_resources = {}
        dom = self.parse_frame_dom_resources(data)
        return self.prepare_rg_requests(self.running_test, dom, self.request_resources)

    def prepare_rg_requests(self, running_test, dom, request_resources):
        # type: (RunningTest, RGridDom, Dict) -> RenderRequest
        if self.size_mode == "region" and self.region_to_check is None:
            raise EyesError("Region to check should be present")
        if self.size_mode == "selector" and not isinstance(
            self.selector, VisualGridSelector
        ):
            raise EyesError("Selector should be present")

        region = None
        if self.region_to_check:
            region = dict(
                x=self.region_to_check.x,
                y=self.region_to_check.y,
                width=self.region_to_check.width,
                height=self.region_to_check.height,
            )
        r_info = RenderInfo(
            width=running_test.browser_info.width,
            height=running_test.browser_info.height,
            size_mode=self.size_mode,
            selector=self.selector,
            region=region,
            emulation_info=running_test.browser_info.emulation_info,
        )
        return RenderRequest(
            webhook=self.rendering_info.results_url,
            agent_id=self.agent_id,
            url=dom.url,
            dom=dom,
            resources=request_resources,
            render_info=r_info,
            browser_name=running_test.browser_info.browser_type,
            platform=running_test.browser_info.platform,
            script_hooks=self.script_hooks,
            selectors_to_find_regions_for=list(chain(*self.region_selectors)),
            send_dom=running_test.configuration.send_dom,
        )

    def parse_frame_dom_resources(self, data):
        # type: (Dict) -> RGridDom
        base_url = data["url"]
        resource_urls = data.get("resourceUrls", [])
        blobs = data.get("blobs", [])
        frames = data.get("frames", [])
        discovered_resources_urls = []

        def handle_resources(content_type, content):
            if content_type == "text/css":
                urls_from_css = css_parser.get_urls_from_css_resource(content)
                for discovered_url in urls_from_css:
                    target_url = _apply_base_url(discovered_url, base_url)
                    with self.discovered_resources_lock:
                        discovered_resources_urls.append(target_url)

        def get_resource(link):
            # type: (Text) -> VGResource
            response = self.eyes_connector.download_resource(link)
            return VGResource.from_response(link, response, on_created=handle_resources)

        for f_data in frames:
            f_data["url"] = _apply_base_url(f_data["url"], base_url)
            self.request_resources[f_data["url"]] = self.parse_frame_dom_resources(
                f_data
            ).resource

        for blob in blobs:
            resource = VGResource.from_blob(blob, on_created=handle_resources)
            if resource.url.rstrip("#") == base_url:
                continue

            self.all_blobs.append(resource)
            self.request_resources[resource.url] = resource

        for r_url in set(resource_urls):
            self.resource_cache.fetch_and_store(r_url, get_resource)
        for r_url in discovered_resources_urls:
            self.resource_cache.fetch_and_store(r_url, get_resource)
        self.request_resources.update(self.resource_cache)
        return RGridDom(
            url=base_url, dom_nodes=data["cdt"], resources=self.request_resources
        )

    def poll_render_status(self, render_request):
        # type: (RenderRequest) -> List[RenderStatusResults]
        iterations = 0
        statuses = []  # type: List[RenderStatusResults]
        if not render_request.render_id:
            raise EyesError("RenderStatus: Got empty renderId!")
        fails_count = 0
        finished = False
        while True:
            if finished:
                break
            while True:
                try:
                    statuses = self.eyes_connector.render_status_by_id(
                        render_request.render_id
                    )
                except Exception as e:
                    logger.exception(e)
                    datetime_utils.sleep(1000)
                    fails_count += 1
                finally:
                    iterations += 1
                    datetime_utils.sleep(500)
                if statuses or 0 < fails_count < 3:
                    break
            finished = bool(
                statuses
                and statuses[0] is not None
                and (
                    statuses[0].status != RenderStatus.RENDERING
                    or iterations > self.MAX_ITERATIONS
                    or False
                )
            )
        if statuses[0].status == RenderStatus.ERROR:
            raise EyesError(
                "Got error during rendering: \n\t{}".format(statuses[0].error)
            )
        return statuses


def _apply_base_url(discovered_url, base_url):
    url = urlparse(discovered_url)
    if url.scheme in ["http", "https"] and url.netloc:
        return discovered_url
    return urljoin(base_url, discovered_url)



import collections
import typing

import attr

from applitools.common import RectangleSize, Region, logger
from applitools.common.utils import image_utils
from applitools.core.capture import ImageProvider
from applitools.selenium.positioning import ScrollPositionProvider
from applitools.selenium.useragent import BrowserNames, OSNames, UserAgent

if typing.TYPE_CHECKING:
    from PIL.Image import Image
    from applitools.selenium.selenium_eyes import SeleniumEyes
    from applitools.selenium.webdriver import EyesWebDriver
    from applitools.selenium.frames import FrameChain


def get_image_provider(ua, eyes):
    # type: (UserAgent, SeleniumEyes) -> ImageProvider
    if ua:
        if ua.browser == BrowserNames.Firefox:
            if ua.browser_major_version >= 48:
                return FirefoxScreenshotImageProvider(eyes)
        elif ua.browser == BrowserNames.Safari:
            return SafariScreenshotImageProvider(eyes, ua)
        elif ua.browser == BrowserNames.MobileSafari:
            return MobileSafariScreenshotImageProvider(eyes, ua)
    return TakesScreenshotImageProvider(eyes)


@attr.s(hash=False)
class DeviceData(object):
    width = attr.ib()
    height = attr.ib()
    vp_width = attr.ib()
    vp_height = attr.ib()

    def __hash__(self):
        return int(
            self.width * 100000
            + self.height * 1000
            + self.vp_width * 100
            + self.vp_height
        )


@attr.s
class MobileSafariScreenshotImageProvider(ImageProvider):
    _devices_regions = None

    _eyes = attr.ib()  # type: SeleniumEyes
    _useragent = attr.ib()  # type: UserAgent

    def get_image(self):
        image = TakesScreenshotImageProvider(self._eyes).get_image()

        if self._eyes.is_cut_provider_explicitly_set:
            return image

        # original_viewport_size = self._eyes._get_viewport_size()
        # FIXME: huck to get correct viewport size on mobile
        original_viewport_size = RectangleSize.from_(
            self._eyes.scroll_root_element.size_and_borders.size
        )
        logger.info("logical viewport size: {}".format(original_viewport_size))

        if self._devices_regions is None:
            self.init_device_regions_table()

        logger.info(
            "physical device pixel size: {} x {}".format(image.width, image.height)
        )

        device_data = DeviceData(
            image.width,
            image.height,
            original_viewport_size.width,
            original_viewport_size.height,
        )
        if device_data in self._devices_regions:
            logger.debug("device model found in hash table")
            crop_by_version = self._devices_regions.get(device_data)
            if crop_by_version.major_version <= self._useragent.browser_major_version:
                image = image_utils.crop_image(image, crop_by_version.region)
            else:
                logger.debug(
                    "device version not found in list. returning original image."
                )
        else:
            logger.debug("device not found in list. returning original image.")
        return image

    def init_device_regions_table(self):
        r_v = collections.namedtuple("RegionAndVersion", "major_version region")
        self._devices_regions = {
            DeviceData(828, 1792, 414, 719): r_v(12, Region(0, 189, 828, 1436)),
            DeviceData(1792, 828, 808, 364): r_v(12, Region(88, 101, 1616, 685)),
            DeviceData(1242, 2688, 414, 719): r_v(12, Region(0, 283, 1242, 2155)),
            DeviceData(2688, 1242, 808, 364): r_v(12, Region(132, 151, 2424, 1028)),
            DeviceData(1125, 2436, 375, 635): r_v(11, Region(0, 283, 1125, 1903)),
            DeviceData(2436, 1125, 724, 325): r_v(11, Region(132, 151, 2436, 930)),
            DeviceData(1242, 2208, 414, 622): r_v(11, Region(0, 211, 1242, 1863)),
            DeviceData(2208, 1242, 736, 364): r_v(11, Region(0, 151, 2208, 1090)),
            DeviceData(1242, 2208, 414, 628): r_v(10, Region(0, 193, 1242, 1882)),
            DeviceData(2208, 1242, 736, 337): r_v(10, Region(0, 231, 2208, 1010)),
            DeviceData(750, 1334, 375, 553): r_v(11, Region(0, 141, 750, 1104)),
            DeviceData(1334, 750, 667, 325): r_v(11, Region(0, 101, 1334, 648)),
            DeviceData(750, 1334, 375, 559): r_v(10, Region(0, 129, 750, 1116)),
            DeviceData(1334, 750, 667, 331): r_v(10, Region(0, 89, 1334, 660)),
            DeviceData(640, 1136, 320, 460): r_v(10, Region(0, 129, 640, 918)),
            DeviceData(1136, 640, 568, 232): r_v(10, Region(0, 89, 1136, 462)),
            DeviceData(1536, 2048, 768, 954): r_v(11, Region(0, 141, 1536, 1907)),
            DeviceData(2048, 1536, 1024, 698): r_v(11, Region(0, 141, 2048, 1395)),
            DeviceData(1536, 2048, 768, 922): r_v(11, Region(0, 206, 1536, 1842)),
            DeviceData(2048, 1536, 1024, 666): r_v(11, Region(0, 206, 2048, 1330)),
            DeviceData(1536, 2048, 768, 960): r_v(10, Region(0, 129, 1536, 1919)),
            DeviceData(2048, 1536, 1024, 704): r_v(10, Region(0, 129, 2048, 1407)),
            DeviceData(1536, 2048, 768, 928): r_v(10, Region(0, 194, 1536, 1854)),
            DeviceData(2048, 1536, 1024, 672): r_v(10, Region(0, 194, 2048, 1342)),
            DeviceData(2048, 2732, 1024, 1296): r_v(11, Region(0, 141, 2048, 2591)),
            DeviceData(2732, 2048, 1366, 954): r_v(11, Region(0, 141, 2732, 1907)),
            DeviceData(1668, 2224, 834, 1042): r_v(11, Region(0, 141, 1668, 2083)),
            DeviceData(2224, 1668, 1112, 764): r_v(11, Region(0, 141, 2224, 1527)),
        }


@attr.s
class SafariScreenshotImageProvider(ImageProvider):

    _eyes = attr.ib()  # type: SeleniumEyes
    _useragent = attr.ib()  # type: UserAgent

    def get_image(self):
        image = TakesScreenshotImageProvider(self._eyes).get_image()

        if self._eyes.is_cut_provider_explicitly_set:
            return image

        scale_ratio = self._eyes.device_pixel_ratio
        original_viewport_size = self._eyes._get_viewport_size()
        viewport_size = original_viewport_size.scale(scale_ratio)

        logger.info("logical viewport size: {}".format(original_viewport_size))

        force_full_page_screenshot = self._eyes.configuration.force_full_page_screenshot
        if force_full_page_screenshot is not None:
            if not force_full_page_screenshot:
                current_frame_chain = self._eyes.driver.frame_chain  # type: FrameChain

                if len(current_frame_chain) == 0:
                    position_provider = ScrollPositionProvider(
                        self._eyes.driver,
                        self._eyes.driver.find_element_by_tag_name("html"),
                    )
                    loc = position_provider.get_current_position()
                else:
                    loc = current_frame_chain.default_content_scroll_position

                loc = loc.scale(scale_ratio)
                image = image_utils.crop_image(image, Region.from_(loc, viewport_size))
        return image


class TakesScreenshotImageProvider(ImageProvider):
    def get_image(self):
        # type: () -> Image
        logger.info("Getting screenshot as base64...")
        screenshot64 = self._eyes.driver.get_screenshot_as_base64()
        logger.info("Done getting base64! Creating BufferedImage...")
        image = image_utils.image_from_base64(screenshot64)
        self._eyes.debug_screenshot_provider.save(image, "initial")
        return image


class FirefoxScreenshotImageProvider(ImageProvider):
    def get_image(self):
        driver = self._eyes.driver  # type: EyesWebDriver
        fc = driver.frame_chain.clone()
        logger.info("frameChain size: {}".format(fc.size))
        logger.info("Switching temporarily to default content.")
        driver.switch_to.default_content()

        logger.info("Getting screenshot as base64.")
        screenshot64 = driver.get_screenshot_as_base64()
        logger.info("Done getting base64! Creating BufferedImage...")

        image = image_utils.image_from_base64(screenshot64)
        logger.info("Done. Switching back to original frame.")
        driver.switch_to.frames(fc)
        return image



import math
import typing
from abc import abstractmethod

import attr

from applitools.common import logger
from applitools.common.utils import ABC
from applitools.selenium.useragent import BrowserNames, OSNames

if typing.TYPE_CHECKING:
    from applitools.common import Region
    from .webdriver import EyesWebDriver
    from .eyes import Eyes
    from .useragent import UserAgent


def get_region_position_compensation(useragent, eyes):
    # type: (UserAgent, Eyes) -> RegionPositionCompensation
    if useragent:
        if useragent.browser == BrowserNames.Firefox:
            if useragent.browser_major_version >= 48:
                return FirefoxRegionPositionCompensation(eyes, useragent)
        elif useragent.browser == BrowserNames.Safari:
            return SafariRegionPositionCompensation(eyes, useragent)
        elif useragent.browser == BrowserNames.IE:
            return InternetExplorerRegionPositionCompensation(eyes, useragent)
    return NullRegionPositionCompensation(eyes, useragent)


@attr.s
class RegionPositionCompensation(ABC):
    _eyes = attr.ib()  # type: Eyes
    _useragent = attr.ib()  # type: UserAgent

    @abstractmethod
    def compensate_region_position(self, region, pixel_ratio):
        pass


class NullRegionPositionCompensation(RegionPositionCompensation):
    def compensate_region_position(self, region, pixel_ratio):
        # type: (Region, float) -> Region
        return region


class FirefoxRegionPositionCompensation(RegionPositionCompensation):
    def compensate_region_position(self, region, pixel_ratio):
        logger.info(str(self._useragent))
        logger.info(pixel_ratio)

        if (
            self._useragent.os == OSNames.Windows
            and self._useragent.os_major_version <= 7
        ):
            logger.info("compensating by {} pixels".format(pixel_ratio))
            return region.offset(0, pixel_ratio)

        if pixel_ratio == 1.0:
            return region

        if self._useragent.browser_major_version > 60:
            return region

        driver = self._eyes.driver  # type: EyesWebDriver
        fc = driver.frame_chain
        logger.info("frame_chain size: {}".format(fc.size))
        if fc.size > 0:
            return region

        region = region.offset(0, -math.ceil(pixel_ratio / 2))
        if region.width <= 0 or region.height <= 0:
            return Region.EMPTY()

        return region


class SafariRegionPositionCompensation(RegionPositionCompensation):
    def compensate_region_position(self, region, pixel_ratio):
        if pixel_ratio == 1.0:
            return region
        if region.width <= 0 and region.height <= 0:
            return Region.EMPTY()
        return region.offset(0, math.ceil(pixel_ratio))


class InternetExplorerRegionPositionCompensation(RegionPositionCompensation):
    def compensate_region_position(self, region, pixel_ratio):
        return region.offset(0, math.ceil(pixel_ratio))



from __future__ import unicode_literals

from typing import TYPE_CHECKING, List, Optional, Text, Tuple, Union, overload

import attr
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

from applitools.common import Region, logger
from applitools.common.utils.compat import basestring
from applitools.core.fluent import CheckSettings, CheckSettingsValues
from applitools.selenium.fluent import SelectorByElement, SelectorByLocator
from applitools.selenium.webelement import EyesWebElement

from .region import (
    FloatingRegionByCssSelector,
    FloatingRegionByElement,
    RegionByCssSelector,
    RegionByElement,
)

if TYPE_CHECKING:
    from applitools.common.visual_grid import VisualGridSelector
    from applitools.common.utils.custom_types import (
        FrameReference,
        AnyWebElement,
        FrameNameOrId,
        FrameIndex,
        BySelector,
        CssSelector,
        FLOATING_VALUES,
    )

BEFORE_CAPTURE_SCREENSHOT = "beforeCaptureScreenshot"


@attr.s
class FrameLocator(object):
    frame_element = attr.ib(default=None)  # type: AnyWebElement
    frame_selector = attr.ib(default=None)  # type: CssSelector
    frame_name_or_id = attr.ib(default=None)  # type: FrameNameOrId
    frame_index = attr.ib(default=None)  # type: FrameIndex
    scroll_root_selector = attr.ib(default=None)  # type: CssSelector
    scroll_root_element = attr.ib(default=None)  # type: AnyWebElement


@attr.s
class SeleniumCheckSettingsValues(CheckSettingsValues):
    # hide_caret = attr.ib(init=False, default=None)
    scroll_root_element = attr.ib(init=False, default=None)  # type: EyesWebElement
    scroll_root_selector = attr.ib(init=False, default=None)  # type: CssSelector
    target_selector = attr.ib(init=False, default=None)  # type: CssSelector
    target_element = attr.ib(init=False, default=None)  # type: EyesWebElement
    frame_chain = attr.ib(init=False, factory=list)  # type: List[FrameLocator]

    # for Rendering Grid
    selector = attr.ib(default=None)  # type: VisualGridSelector
    script_hooks = attr.ib(factory=dict)  # type: dict

    @property
    def target_provider(self):
        target_selector = self.target_selector
        target_element = self.target_element
        if target_selector:
            return SelectorByLocator(target_selector)
        elif target_element:
            return SelectorByElement(target_element)

    @property
    def size_mode(self):
        if (
            self.target_region is None
            and self.target_selector is None
            and self.target_element is None
        ):
            if self.stitch_content:
                return "full-page"
            return "viewport"
        if self.target_region:
            return "region"
        return "selector"


def _css_selector_from_(by, value):
    if by == By.ID:
        value = "#%s" % value
    elif by == By.CLASS_NAME:
        value = ".%s" % value
    elif by == By.NAME:
        value = '[name="%s"]' % value
    elif by in [By.XPATH, By.CSS_SELECTOR, By.TAG_NAME]:
        value = value
    else:
        raise TypeError("By {} is not supported".format(by))
    return value


@attr.s
class SeleniumCheckSettings(CheckSettings):
    values = attr.ib(
        init=False, factory=SeleniumCheckSettingsValues
    )  # type: SeleniumCheckSettingsValues

    _region = attr.ib(default=None)
    _frame = attr.ib(default=None)

    @overload  # noqa
    def layout(self, *by):
        # type: (*BySelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *element):
        # type: (*AnyWebElement)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *css_selector):
        # type: (*CssSelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *region):
        # type: (*Region)  -> SeleniumCheckSettings
        pass

    def layout(self, *region):  # noqa
        return super(SeleniumCheckSettings, self).layout(*region)

    @overload  # noqa
    def strict(self, *by):
        # type: (*BySelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *element):
        # type: (*AnyWebElement)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *css_selector):
        # type: (*CssSelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *region):
        # type: (*Region)  -> SeleniumCheckSettings
        pass

    def strict(self, *region):  # noqa
        return super(SeleniumCheckSettings, self).strict(*region)

    @overload  # noqa
    def content(self, *by):
        # type: (*BySelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *element):
        # type: (*AnyWebElement)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *css_selector):
        # type: (*CssSelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *region):
        # type: (*Region)  -> SeleniumCheckSettings
        pass

    def content(self, *region):  # noqa
        return super(SeleniumCheckSettings, self).content(*region)

    @overload  # noqa
    def ignore(self, *by):
        # type: (*BySelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *element):
        # type: (*AnyWebElement)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *css_selector):
        # type: (*CssSelector)  -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *region):
        # type: (*Region)  -> SeleniumCheckSettings
        pass

    def ignore(self, *region):  # noqa
        return super(SeleniumCheckSettings, self).ignore(*region)

    @overload  # noqa
    def floating(self, max_offset, region):
        # type: (int, FLOATING_VALUES) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def floating(
        self, region, max_up_offset, max_down_offset, max_left_offset, max_right_offset
    ):
        # type: (FLOATING_VALUES, int, int, int, int) -> SeleniumCheckSettings
        pass

    def floating(self, *args):  # noqa
        return super(SeleniumCheckSettings, self).floating(*args)

    @overload  # noqa
    def region(self, region):
        # type: (Region) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, css_selector):
        # type: (CssSelector) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, by):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    def region(self, region):  # noqa
        if isinstance(region, Region):
            self.values.target_region = region
        elif is_list_or_tuple(region):
            by, value = region
            self.values.target_selector = _css_selector_from_(by, value)
        elif isinstance(region, basestring):
            self.values.target_selector = region
        elif is_webelement(region):
            self.values.target_element = region
        else:
            raise TypeError("region method called with argument of unknown type!")
        return self

    @overload  # noqa
    def frame(self, frame_name_or_id):
        # type: (FrameNameOrId) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, index):
        # type: (FrameIndex) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, by):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    def frame(self, frame):  # noqa
        fl = FrameLocator()
        if isinstance(frame, int):
            fl.frame_index = frame
        elif isinstance(frame, basestring):
            fl.frame_name_or_id = frame
        elif is_webelement(frame):
            fl.frame_element = frame
        elif is_list_or_tuple(frame):
            by, value = frame
            selector = _css_selector_from_(by, value)
            fl.frame_selector = selector
        else:
            raise TypeError("frame method called with argument of unknown type!")
        self.values.frame_chain.append(fl)
        return self

    def before_render_screenshot_hook(self, hook):
        # type: (Text) -> SeleniumCheckSettings
        self.values.script_hooks[BEFORE_CAPTURE_SCREENSHOT] = hook
        return self

    def _region_provider_from(self, region, method_name):
        if isinstance(region, basestring):
            logger.debug("{name}: RegionByCssSelector".format(name=method_name))
            return RegionByCssSelector(region)
        if is_list_or_tuple(region):
            by, val = region
            sel = _css_selector_from_(by, val)
            logger.debug("{name}: RegionByCssSelector".format(name=method_name))
            return RegionByCssSelector(sel)
        elif is_webelement(region):
            logger.debug("{name}: RegionByElement".format(name=method_name))
            return RegionByElement(region)
        return super(SeleniumCheckSettings, self)._region_provider_from(
            region, method_name
        )

    def _set_scroll_root_selector(self, selector):
        if len(self.values.frame_chain) == 0:
            self.values.scroll_root_selector = selector
        else:
            self.values.frame_chain[-1].scroll_root_selector = selector

    def _set_scroll_root_element(self, element):
        if len(self.values.frame_chain) == 0:
            self.values.scroll_root_element = element
        else:
            self.values.frame_chain[-1].scroll_root_element = element

    @overload  # noqa
    def scroll_root_element(self, element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def scroll_root_element(self, selector):
        # type: (CssSelector) -> SeleniumCheckSettings
        pass

    def scroll_root_element(self, element_or_selector):  # noqa
        if isinstance(element_or_selector, basestring):
            self._set_scroll_root_selector(element_or_selector)
        elif is_webelement(element_or_selector):
            self._set_scroll_root_element(element_or_selector)
        return self

    def _floating_provider_from(self, region, bounds):
        if is_webelement(region):
            logger.debug("floating: FloatingRegionByElement")
            return FloatingRegionByElement(region, bounds)
        if isinstance(region, basestring):
            logger.debug("floating: FloatingRegionByCssSelector")
            return FloatingRegionByCssSelector(region, bounds)
        if is_list_or_tuple(region):
            by, value = region
            selector = _css_selector_from_(by, value)
            logger.debug("floating: FloatingRegionByCssSelector")
            return FloatingRegionByCssSelector(selector, bounds)
        return super(SeleniumCheckSettings, self)._floating_provider_from(
            region, bounds
        )


def is_list_or_tuple(elm):
    return isinstance(elm, list) or isinstance(elm, tuple)


def is_webelement(elm):
    return (
        isinstance(elm, EyesWebElement)
        or isinstance(elm, WebElement)
        or isinstance(getattr(elm, "_element", None), WebElement)
    )



from .region import (
    FloatingRegionBy,
    FloatingRegionByCssSelector,
    FloatingRegionByElement,
    RegionBy,
    RegionByCssSelector,
    RegionByElement,
)
from .selector import SelectorByElement, SelectorByLocator
from .selenium_check_settings import FrameLocator, SeleniumCheckSettings
from .target import Target

__all__ = (
    "Target",
    "SeleniumCheckSettings",
    "FrameLocator",
    "RegionBy",
    "RegionByElement",
    "RegionByCssSelector",
    "FloatingRegionBy",
    "FloatingRegionByElement",
    "FloatingRegionByCssSelector",
    "SelectorByElement",
    "SelectorByLocator",
)



from __future__ import absolute_import

import math
import typing as tp

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By

from applitools.common import logger
from applitools.common.geometry import CoordinatesType, Point, Region
from applitools.common.utils.general_utils import proxy_to

from . import eyes_selenium_utils

if tp.TYPE_CHECKING:
    from typing import Optional, Text
    from selenium.webdriver.remote.webelement import WebElement
    from .positioning import SeleniumPositionProvider
    from .webdriver import EyesWebDriver

_JS_GET_COMPUTED_STYLE_FORMATTED_STR = """
            var elem = arguments[0];
            var styleProp = '%s';
            if (window.getComputedStyle) {
                return window.getComputedStyle(elem, null)
                .getPropertyValue(styleProp);
            } else if (elem.currentStyle) {
                return elem.currentStyle[styleProp];
            } else {
                return null;
            }
    """
_JS_GET_SCROLL_LEFT = "return arguments[0].scrollLeft;"
_JS_GET_SCROLL_TOP = "return arguments[0].scrollTop;"
_JS_GET_SCROLL_WIDTH = "return arguments[0].scrollWidth;"
_JS_GET_SCROLL_HEIGHT = "return arguments[0].scrollHeight;"
_JS_GET_OVERFLOW = "return arguments[0].style.overflow;"
_JS_SET_OVERFLOW_FORMATTED_STR = "arguments[0].style.overflow = '%s'"
_JS_GET_CLIENT_WIDTH = "return arguments[0].clientWidth;"
_JS_GET_CLIENT_HEIGHT = "return arguments[0].clientHeight;"
_JS_SCROLL_TO_FORMATTED_STR = """
        arguments[0].scrollLeft = {:d};
        arguments[0].scrollTop = {:d};
"""
_JS_GET_SCROLL_POSITION = """
return arguments[0].scrollLeft + ';' + arguments[0].scrollTop;
"""
_JS_GET_BORDER_WIDTHS_ARR = """
        var retVal = retVal || [];
        if (window.getComputedStyle) {
        var computedStyle = window.getComputedStyle(elem, null);
        retVal.push(computedStyle.getPropertyValue('border-left-width'));
        retVal.push(computedStyle.getPropertyValue('border-top-width'));
        retVal.push(computedStyle.getPropertyValue('border-right-width'));
        retVal.push(computedStyle.getPropertyValue('border-bottom-width'));
        } else if (elem.currentStyle) {
        retVal.push(elem.currentStyle['border-left-width']);
        retVal.push(elem.currentStyle['border-top-width']);
        retVal.push(elem.currentStyle['border-right-width']);
        retVal.push(elem.currentStyle['border-bottom-width']);
        } else {
        retVal.push(0,0,0,0);
        }
"""
_JS_GET_BORDER_WIDTHS = _JS_GET_BORDER_WIDTHS_ARR + "return retVal;"
_JS_GET_SIZE_AND_BORDER_WIDTHS = """
        var elem = arguments[0];
        var retVal = [arguments[0].clientWidth, arguments[0].clientHeight];
        {}
        return retVal
""".format(
    _JS_GET_BORDER_WIDTHS_ARR
)


@proxy_to(
    "_element",
    [
        "tag_name",
        "text",
        "submit",
        "clear",
        "get_property",
        "get_attribute",
        "is_selected",
        "is_enabled",
        "find_element_by_id",
        "find_elements_by_id",
        "find_element_by_name",
        "find_elements_by_name",
        "find_element_by_link_text",
        "find_elements_by_link_text",
        "find_element_by_partial_link_text",
        "find_elements_by_partial_link_text",
        "find_element_by_tag_name",
        "find_elements_by_tag_name",
        "find_element_by_xpath",
        "find_elements_by_xpath",
        "find_element_by_class_name",
        "find_elements_by_class_name",
        "find_element_by_css_selector",
        "find_elements_by_css_selector",
        "is_displayed",
        "location_once_scrolled_into_view",
        "parent",
        "screenshot_as_base64",
        "screenshot_as_png",
        "screenshot",
        "location_in_view",
        "anonymous_children",
    ],
)
class EyesWebElement(object):
    """
    A wrapper for selenium web element. This enables eyes to be notified
     about actions/events for this element.
    """

    def __init__(self, element, driver):
        # type: (WebElement, EyesWebDriver) -> None
        """
        Ctor.

        :param element: The element in the frame.
        :param eyes: The eyes sdk instance.
        :param driver: EyesWebDriver instance.
        """
        if isinstance(element, EyesWebElement):
            element = element._element
        self._element = element  # type: WebElement
        self._driver = driver  # type: EyesWebDriver

        # setting from outside
        self.position_provider = None  # type: Optional[SeleniumPositionProvider]
        self._original_overflow = None  # type: Optional[Text]

    @property
    def element(self):
        # type: () -> WebElement
        return self._element

    @property
    def driver(self):
        # type: () -> EyesWebDriver
        return self._driver

    @property
    def bounds(self):
        # type: () -> Region
        # noinspection PyUnresolvedReferences
        location = self.location
        left, top = location["x"], location["y"]
        width = height = 0  # Default

        # noinspection PyBroadException
        try:
            size = self._element.size
            width, height = size["width"], size["height"]
        except Exception:
            # Not implemented on all platforms.
            pass
        if left < 0:
            left, width = 0, max(0, width + left)
        if top < 0:
            top, height = 0, max(0, height + top)
        return Region(left, top, width, height, CoordinatesType.CONTEXT_RELATIVE)

    @property
    def location(self):
        return self._element.location

    @property
    def rect(self):
        return self._element.rect

    @property
    def is_attached_to_page(self):
        try:
            self.location
            return True
        except WebDriverException:
            return False

    @property
    def size(self):
        return self._element.size

    @property
    def scroll_location(self):
        # type: () -> Point
        pos = self.driver.execute_script(_JS_GET_SCROLL_POSITION, self.element)
        return eyes_selenium_utils.parse_location_string(pos)

    @property
    def id(self):
        return self._element.id

    def value_of_css_property(self, property_name):
        return self._element.value_of_css_property(property_name)

    def find_element(self, by=By.ID, value=None):
        """
        Returns a WebElement denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: WebElement denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        result = self._original_methods["find_element"](by, value)
        # Wrap the element.
        if result:
            result = EyesWebElement(result, self._driver)
        return result

    def find_elements(self, by=By.ID, value=None):
        """
        Returns a list of web elements denoted by "By".

        :param by: By which option to search for (default is by ID).
        :param value: The value to search for.
        :return: List of web elements denoted by "By".
        """
        # Get result from the original implementation of the underlying driver.
        results = self._original_methods["find_elements"](by, value)
        # Wrap all returned elements.
        if results:
            updated_results = []
            for element in results:
                updated_results.append(EyesWebElement(element, self._driver))
            results = updated_results
        return results

    def click(self):
        """
        Clicks and element.
        """
        self._driver.eyes.add_mouse_trigger_by_element("click", self)
        self._element.click()

    def send_keys(self, *value):
        """
        Sends keys to a certain element.

        :param value: The value to type into the element.
        """
        text = u""
        for val in value:
            if isinstance(val, int):
                val = val.__str__()
            text += val.encode("utf-8").decode("utf-8")
        self._driver.eyes.add_text_trigger_by_element(self, text)
        self._element.send_keys(*value)

    def hide_scrollbars(self):
        # type: () -> tp.Text
        """
        Hides the scrollbars of the current element.

        :return: The previous value of the overflow property (could be None).
        """
        logger.debug("EyesWebElement.hide_scrollbars()")
        self._original_overflow = eyes_selenium_utils.hide_scrollbars(  # type: ignore
            self.driver, self.element
        )
        return self._original_overflow

    def return_to_original_overflow(self):
        return eyes_selenium_utils.return_to_original_overflow(
            self.driver, self.element, self._original_overflow
        )

    def get_computed_style(self, prop_style):
        script = _JS_GET_COMPUTED_STYLE_FORMATTED_STR % prop_style
        return self._driver.execute_script(script, self._element)

    def get_computed_style_int(self, prop_style):

        value = self.get_computed_style(prop_style)
        return int(round(float(value.replace("px", "").strip())))

    def get_scroll_left(self):
        return int(
            math.ceil(self._driver.execute_script(_JS_GET_SCROLL_LEFT, self._element))
        )

    def get_scroll_top(self):
        return math.ceil(self._driver.execute_script(_JS_GET_SCROLL_TOP, self._element))

    def get_scroll_width(self):
        return int(
            math.ceil(self._driver.execute_script(_JS_GET_SCROLL_WIDTH, self._element))
        )

    def get_scroll_height(self):
        return int(
            math.ceil(self._driver.execute_script(_JS_GET_SCROLL_HEIGHT, self._element))
        )

    def get_border_left_width(self):
        return self.get_computed_style_int("border-left-width")

    def get_border_right_width(self):
        return self.get_computed_style_int("border-right-width")

    def get_border_top_width(self):
        return self.get_computed_style_int("border-right-width")

    def get_border_bottom_width(self):
        return self.get_computed_style_int("border-right-width")

    def get_overflow(self):
        return self._driver.execute_script(_JS_GET_OVERFLOW, self._element)

    def get_client_width(self):
        return int(
            math.ceil(
                float(self._driver.execute_script(_JS_GET_CLIENT_WIDTH, self._element))
            )
        )

    def get_client_height(self):
        return int(
            math.ceil(
                float(self._driver.execute_script(_JS_GET_CLIENT_HEIGHT, self._element))
            )
        )

    def scroll_to(self, location):
        # type: (Point) -> Point
        """Scrolls to the specified location inside the element."""
        position = self._driver.execute_script(
            _JS_SCROLL_TO_FORMATTED_STR.format(location["x"], location["y"])
            + _JS_GET_SCROLL_POSITION,
            self._element,
        )
        return eyes_selenium_utils.parse_location_string(position)

    @property
    def size_and_borders(self):
        # type: () -> SizeAndBorders
        ret_val = self._driver.execute_script(
            _JS_GET_SIZE_AND_BORDER_WIDTHS, self._element
        )
        return SizeAndBorders(
            width=int(ret_val[0]),
            height=int(ret_val[1]),
            left=int(ret_val[2].rstrip("px")),
            top=int(ret_val[3].rstrip("px")),
            right=int(ret_val[4].rstrip("px")),
            bottom=int(ret_val[5].rstrip("px")),
        )

    @property
    def bounding_client_rect(self):
        left, top, width, height = self.driver.execute_script(
            "var r = arguments[0].getBoundingClientRect();"
            "return r.left+';'+r.top+';'+r.width+';'+r.height;",
            self._element,
        ).split(";")
        return dict(
            x=math.ceil(float(left)),
            y=math.ceil(float(top)),
            width=math.ceil(float(width)),
            height=math.ceil(float(height)),
        )

    def __str__(self):
        tag_name = "None"
        if self.is_attached_to_page:
            tag_name = self._element.tag_name
        return "EyesWebElement: id {}, tag_name {}".format(self._element.id, tag_name)


class SizeAndBorders(object):
    __slots__ = ("size", "borders")

    def __init__(self, width, height, left, top, right, bottom):
        self.size = dict(width=width, height=height)
        self.borders = dict(left=left, top=top, right=right, bottom=bottom)

    def __str__(self):
        return "SizeAndBorders(size={}, borders={})".format(self.size, self.borders)



from __future__ import absolute_import

import typing

from applitools.common import EyesError, logger
from applitools.common.selenium import Configuration
from applitools.common.utils import argument_guard
from applitools.common.utils.general_utils import all_fields, proxy_to
from applitools.selenium import ClassicRunner, eyes_selenium_utils

from .fluent import Target
from .selenium_eyes import SeleniumEyes
from .visual_grid import VisualGridEyes, VisualGridRunner
from .webdriver import EyesWebDriver

if typing.TYPE_CHECKING:
    from typing import Text, Optional, Union, List, Tuple
    from selenium.webdriver.remote.webelement import WebElement
    from applitools.common import MatchResult, TestResults, Region, SessionType
    from applitools.common.utils.custom_types import (
        AnyWebDriver,
        ViewPort,
        FrameReference,
        AnyWebElement,
    )
    from applitools.core import (
        PositionProvider,
        FixedCutProvider,
        UnscaledFixedCutProvider,
        NullCutProvider,
    )
    from .frames import FrameChain
    from .fluent import SeleniumCheckSettings
    from .webelement import EyesWebElement


@proxy_to("configuration", all_fields(Configuration))
class Eyes(object):
    _is_visual_grid_eyes = False  # type: bool
    _visual_grid_eyes = None  # type: VisualGridEyes
    _selenium_eyes = None  # type: SeleniumEyes
    _runner = None  # type: Optional[VisualGridRunner]
    _driver = None  # type: Optional[EyesWebDriver]
    _is_opened = False  # type: bool

    def __init__(self, runner=None):
        # type: (Optional[VisualGridRunner, ClassicRunner]) -> None
        self._configuration = Configuration()  # type: Configuration

        # backward compatibility with settings server_url
        if isinstance(runner, str):
            self.configuration.server_url = runner
            runner = None

        if runner is None:
            self._selenium_eyes = SeleniumEyes(self, None)
        elif isinstance(runner, VisualGridRunner):
            self._runner = runner
            self._visual_grid_eyes = VisualGridEyes(self, runner)
            self._is_visual_grid_eyes = True
        elif isinstance(runner, ClassicRunner):
            self._runner = runner
            self._selenium_eyes = SeleniumEyes(self, runner)
            self._is_visual_grid_eyes = False
        else:
            raise ValueError("Wrong runner")

    @property
    def is_open(self):
        # type: () -> bool
        return self._is_opened

    def get_configuration(self):
        # type: () -> Configuration
        return self._configuration

    def set_configuration(self, configuration):
        # type: (Configuration) -> None
        argument_guard.is_a(configuration, Configuration)
        if self._configuration.api_key and not configuration.api_key:
            configuration.api_key = self._configuration.api_key
        if self._configuration.server_url and not configuration.server_url:
            configuration.server_url = self._configuration.server_url
        self._configuration = configuration

    configuration = property(get_configuration, set_configuration)

    @property
    def rotation(self):
        # type: () -> Optional[int]
        if self._selenium_eyes and self.driver:
            return self.driver.rotation
        return None

    @rotation.setter
    def rotation(self, rotation):
        # type: (int) -> None
        argument_guard.is_a(rotation, int)
        if self._selenium_eyes and self.driver:
            self.driver.rotation = rotation

    @property
    def base_agent_id(self):
        # type: () -> Text
        """
        Must return version of SDK. (e.g. selenium, visualgrid) in next format:
            "eyes.{package}.python/{lib_version}"
        """
        return self._current_eyes.base_agent_id

    @property
    def full_agent_id(self):
        # type: () -> Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        """
        return self._current_eyes.full_agent_id

    @property
    def should_stitch_content(self):
        # type: () -> bool
        return self._current_eyes.should_stitch_content

    @property
    def original_fc(self):
        # type: () -> Optional[FrameChain]
        """ Gets original frame chain

        Before check() call we save original frame chain

        Returns:
            Frame chain saved before check() call
        """
        return self._current_eyes.original_fc

    # def rotation(self):
    #     if not self._is_visual_grid_eyes:
    #         return self._selenium_eyes.rotation

    @property
    def device_pixel_ratio(self):
        # type: () -> int
        """
        Gets device pixel ratio.

        :return The device pixel ratio, or if the DPR is not known yet or if it wasn't
        possible to extract it.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.device_pixel_ratio
        return 0

    @property
    def scale_ratio(self):
        # type: () -> float
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.scale_ratio
        return 0

    @scale_ratio.setter
    def scale_ratio(self, value):
        # type: (float) -> None
        """
        Manually set the scale ratio for the images being validated.
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.scale_ratio = value

    @property
    def position_provider(self):
        """
        Sets position provider.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.position_provider
        return None

    @property
    def _debug_screenshot_provided(self):
        # type: () -> bool
        """True if screenshots saving enabled."""
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes._debug_screenshot_provided

    @_debug_screenshot_provided.setter
    def _debug_screenshot_provided(self, save):
        # type: (bool) -> None
        if not self._is_visual_grid_eyes:
            self._selenium_eyes._debug_screenshot_provided = save

    @position_provider.setter
    def position_provider(self, provider):
        # type: (PositionProvider) -> None
        """
        Gets position provider.
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.position_provider = provider

    @property
    def cut_provider(self):
        # type:()->Optional[Union[FixedCutProvider,UnscaledFixedCutProvider,NullCutProvider]]
        """
        Gets current cut provider
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.cut_provider
        return None

    @cut_provider.setter
    def cut_provider(self, cutprovider):
        # type: (Union[FixedCutProvider,UnscaledFixedCutProvider,NullCutProvider])->None
        """
        Manually set the the sizes to cut from an image before it's validated.

        :param provider:
        :return:
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.cut_provider = cutprovider

    @property
    def is_cut_provider_explicitly_set(self):
        """
        Gets is cut provider explicitly set.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.is_cut_provider_explicitly_set
        return False

    @property
    def agent_setup(self):
        """
        Gets agent setup.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.agent_setup
        return None

    @property
    def current_frame_position_provider(self):
        # type: () -> Optional[PositionProvider]
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.current_frame_position_provider
        return None

    @staticmethod
    def get_viewport_size(driver):
        # type: (AnyWebDriver) -> ViewPort
        return eyes_selenium_utils.get_viewport_size_or_display_size(driver)

    @staticmethod
    def set_viewport_size(driver, size):
        # type: (AnyWebDriver, ViewPort) -> None
        assert driver is not None
        if size is None:
            raise ValueError("set_viewport_size require `size` parameter")
        eyes_selenium_utils.set_viewport_size(driver, size)

    def add_property(self, name, value):
        # type: (Text, Text) -> None
        """
        Associates a key/value pair with the test. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self._current_eyes.add_property(name, value)

    def clear_properties(self):
        """
        Clears the list of custom properties.
        """
        self._current_eyes.clear_properties()

    def add_mouse_trigger_by_element(self, action, element):
        # type: (Text, AnyWebElement) -> None
        """
        Adds a mouse trigger.

        :param action: Mouse action (click, double click etc.)
        :param element: The element on which the action was performed.
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.add_mouse_trigger_by_element(action, element)

    def add_text_trigger_by_element(self, element, text):
        # type: (AnyWebElement, Text) -> None
        """
        Adds a text trigger.

        :param element: The element to which the text was sent.
        :param text: The trigger's text.
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.add_text_trigger_by_element(element, text)

    @property
    def driver(self):
        # type: () -> EyesWebDriver
        return self._driver

    @property
    def send_dom(self):
        # type: () -> bool
        if not self._is_visual_grid_eyes:
            return self.configuration.send_dom
        return False

    @send_dom.setter
    def send_dom(self, value):
        # type: (bool) -> None
        if not self._is_visual_grid_eyes:
            self.configuration.send_dom = value

    def check(self, name, check_settings):
        # type: (Text, SeleniumCheckSettings) -> MatchResult
        """
        Takes a snapshot and matches it with the expected output.

        :param name: The name of the tag.
        :param check_settings: target which area of the window to check.
        :return: The match results.
        """
        if self.configuration.is_disabled:
            return MatchResult()
        if not self.is_open:
            self.abort()
            raise EyesError("you must call open() before checking")
        return self._current_eyes.check(name, check_settings)

    def check_window(self, tag=None, match_timeout=-1, fully=None):
        # type: (Optional[Text], int, Optional[bool]) -> MatchResult
        """
        Takes a snapshot of the application under test and matches it with the expected
         output.

        :param tag: An optional tag to be associated with the snapshot.
        :param match_timeout:  The amount of time to retry matching (milliseconds)
        :param fully: Defines that the screenshot will contain the entire window.
        :return: The match results.
        """
        logger.debug("check_window('%s')" % tag)
        return self.check(tag, Target.window().timeout(match_timeout).fully(fully))

    def check_frame(self, frame_reference, tag=None, match_timeout=-1):
        # type: (FrameReference, Optional[Text], int) -> MatchResult
        """
        Check frame.

        :param frame_reference: The name or id of the frame to check. (The same
                name/id as would be used in a call to driver.switch_to.frame()).
        :param tag: An optional tag to be associated with the match.
        :param match_timeout: The amount of time to retry matching. (Milliseconds)
        :return: The match results.
        """
        return self.check(
            tag, Target.frame(frame_reference).fully().timeout(match_timeout)
        )

    def check_region(
        self,
        region,  # type: Union[Region,Text,List,Tuple,WebElement,EyesWebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> MatchResult
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param region: The region which will be visually validated. The coordinates are
                       relative to the viewport of the current frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        :return: The match results.
        """
        return self.check(
            tag,
            Target.region(region).timeout(match_timeout).stitch_content(stitch_content),
        )

    def check_region_in_frame(
        self,
        frame_reference,  # type: FrameReference
        region,  # type: Union[Region,Text,List,Tuple,WebElement,EyesWebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> MatchResult
        """
        Checks a region within a frame, and returns to the current frame.

        :param frame_reference: A reference to the frame in which the region
                                should be checked.
        :param region: Specifying the region to check inside the frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        :return: None
        """
        # TODO: remove this disable
        if self.configuration.is_disabled:
            logger.info("check_region_in_frame_by_selector(): ignored (disabled)")
            return MatchResult()
        logger.debug("check_region_in_frame_by_selector('%s')" % tag)
        return self.check(
            tag,
            Target.region(region)
            .frame(frame_reference)
            .stitch_content(stitch_content)
            .timeout(match_timeout),
        )

    def open(
        self,
        driver,  # type: AnyWebDriver
        app_name=None,  # type: Optional[Text]
        test_name=None,  # type: Optional[Text]
        viewport_size=None,  # type: Optional[ViewPort]
    ):
        # type: (...) -> EyesWebDriver
        """
        Starts a test.

        :param driver: The driver that controls the browser hosting the application
            under the test.
        :param app_name: The name of the application under test.
        :param test_name: The test name.
        :param viewport_size: The client's viewport size (i.e.,
            the visible part of the document's body) or None to allow any viewport size.
        :raise EyesError: If the session was already open.
        """
        if app_name:
            self.configuration.app_name = app_name
        if test_name:
            self.configuration.test_name = test_name
        if viewport_size:
            self.configuration.viewport_size = viewport_size  # type: ignore
        self._init_driver(driver)
        result = self._current_eyes.open(self.driver)
        self._is_opened = True
        return result

    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        result = self._current_eyes.close(raise_ex)
        self._is_opened = False
        return result

    def close_async(self):
        if self._is_visual_grid_eyes:
            self._visual_grid_eyes.close_async()
        else:
            self._selenium_eyes.close(False)

    def abort(self):
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        self._current_eyes.abort()

    def abort_if_not_closed(self):
        logger.deprecation("Use `abort()` instead")
        self.abort()

    def _init_driver(self, driver):
        # type: (AnyWebDriver) -> None
        if isinstance(driver, EyesWebDriver):
            # If the driver is an EyesWebDriver (as might be the case when tests are ran
            # consecutively using the same driver object)
            self._driver = driver
        else:
            self._driver = EyesWebDriver(driver, self)

    @property
    def _current_eyes(self):
        # type: () -> Union[SeleniumEyes, VisualGridEyes]
        if self._is_visual_grid_eyes:
            return self._visual_grid_eyes
        else:
            return self._selenium_eyes

    @property
    def _original_scroll_position(self):
        if self._selenium_eyes:
            return self._selenium_eyes._original_scroll_position
        return None



import typing

import attr
from transitions import Machine

from applitools.common import Region, logger

from .render_task import RenderTask
from .vg_task import VGTask

if typing.TYPE_CHECKING:
    from typing import List, Optional, Dict, Any, Text
    from applitools.common import (
        TestResults,
        VisualGridSelector,
        RenderBrowserInfo,
        RenderStatusResults,
    )
    from applitools.common.selenium import Configuration
    from applitools.selenium.fluent import SeleniumCheckSettings
    from .visual_grid_runner import VisualGridRunner
    from .eyes_connector import EyesConnector

NEW = "new"
NOT_RENDERED = "not_rendered"
RENDERED = "rendered"
OPENED = "opened"
COMPLETED = "completed"
TESTED = "tested"

STATES = [NEW, NOT_RENDERED, OPENED, RENDERED, COMPLETED, TESTED]
TRANSITIONS = [
    {"trigger": "becomes_not_rendered", "source": NEW, "dest": NOT_RENDERED},
    {"trigger": "becomes_opened", "source": RENDERED, "dest": OPENED},
    {"trigger": "becomes_rendered", "source": NOT_RENDERED, "dest": RENDERED},
    {
        "trigger": "becomes_tested",
        "source": [NEW, NOT_RENDERED, OPENED],
        "dest": TESTED,
    },
    {
        "trigger": "becomes_completed",
        "source": [NEW, NOT_RENDERED, RENDERED, OPENED, TESTED],
        "dest": COMPLETED,
    },
]


@attr.s(hash=True, str=False)
class RunningTest(object):
    eyes = attr.ib(hash=False, repr=False)  # type: EyesConnector
    configuration = attr.ib(hash=False, repr=False)  # type: Configuration
    browser_info = attr.ib()  # type: RenderBrowserInfo
    # listener = attr.ib()  # type:
    region_selectors = attr.ib(
        init=False, factory=list, hash=False
    )  # type: List[List[VisualGridSelector]]
    regions = attr.ib(init=False, factory=list, hash=False)

    tasks_list = attr.ib(init=False, factory=list, hash=False)
    task_to_future_mapping = attr.ib(init=False, factory=dict, hash=False)
    test_uuid = attr.ib(init=False)
    on_results = attr.ib(init=False, hash=False)
    state = attr.ib(init=False, hash=False)

    def __str__(self):
        return "RunningTest: state={} uuid={}".format(
            self.state, format(self.test_uuid)
        )

    def __attrs_post_init__(self):
        # type: () -> None
        self._initialize_vars()
        self._initialize_state_machine()
        self.open()

    def _initialize_vars(self):
        # type: () -> None
        self.open_queue = []  # type: List[VGTask]
        self.task_queue = []  # type: List[VGTask]
        self.render_queue = []  # type: List[RenderTask]
        self.close_queue = []  # type: List[VGTask]
        self.watch_open = {}  # type: Dict[VGTask, bool]
        self.watch_task = {}  # type: Dict[VGTask, bool]
        self.watch_render = {}  # type: Dict[RenderTask, bool]
        self.watch_close = {}  # type: Dict[VGTask, bool]
        self.task_lock = None  # type: Optional[VGTask]
        self.test_result = None  # type: Optional[TestResults]
        self.pending_exceptions = []  # type: List[Exception]

    def _initialize_state_machine(self):
        # type: () -> None
        machine = Machine(
            model=self,
            states=STATES,
            transitions=TRANSITIONS,
            initial=NEW,
            send_event=True,
            queued=True,
        )
        self.machine = machine

    def on_results_received(self, func):
        self.on_results = func

    @property
    def queue(self):
        # type: () -> List
        if self.state == NEW:
            return []
        elif self.state == NOT_RENDERED:
            return self.render_queue
        elif self.state == RENDERED:
            return self.open_queue
        elif self.state == OPENED:
            if self.task_lock:
                return []
            elif self.task_queue:
                self.task_lock = self.task_queue[-1]
                return self.task_queue
        elif self.state == TESTED:
            return self.close_queue
        elif self.state == COMPLETED:
            return []
        else:
            raise TypeError("Unsupported state")

    @property
    def score(self):
        # type: () -> int
        if self.state == NEW:
            return 0
        elif self.state == NOT_RENDERED:
            return len(self.render_queue) * 10
        elif self.state == RENDERED:
            return len(self.open_queue)
        elif self.state == OPENED:
            return len(self.task_queue)
        elif self.state == TESTED:
            return len(self.close_queue)
        elif self.state == COMPLETED:
            return 0
        else:
            raise TypeError("Unsupported state")

    def open(self):
        # type: () -> None
        open_task = VGTask(
            "open {}".format(self.browser_info),
            lambda: self.eyes.open(self.configuration),
        )
        logger.debug("RunningTest %s" % open_task.name)

        def open_task_succeeded(test_result):
            # type: (Optional[Any]) -> None
            logger.debug("open_task_succeeded: task.uuid: {}".format(open_task.uuid))
            self.watch_open[open_task] = True
            if self.all_tasks_completed(self.watch_open):
                self.becomes_opened()

        def open_task_error(e):
            logger.debug(
                "open_task_error: task.uuid: {}\n{}".format(open_task.uuid, str(e))
            )
            self.pending_exceptions.append(e)
            self.becomes_completed()

        open_task.on_task_succeeded(open_task_succeeded)
        open_task.on_task_error(open_task_error)
        self.open_queue.append(open_task)
        self.watch_open[open_task] = False

    def check(
        self,
        tag,  # type: Text
        check_settings,  # type: SeleniumCheckSettings
        script_result,  # type: Dict[str, Any]
        visual_grid_manager,  # type: VisualGridRunner
        region_selectors,
        region_to_check,
        script_hooks,
    ):
        # type: (...) -> None
        logger.debug("RunningTest %s , %s" % (tag, check_settings))
        render_task = self._render_task(
            script_result,
            tag,
            visual_grid_manager,
            region_selectors,
            region_to_check,
            script_hooks,
            check_settings,
        )

        def check_run():
            logger.debug("check_run: render_task.uuid: {}".format(render_task.uuid))
            self.eyes.check(
                tag, check_settings, render_task.uuid, region_selectors, self.regions
            )

        check_task = VGTask(
            "perform check {} {}".format(tag, check_settings), check_run
        )

        def check_task_completed():
            # type: () -> None
            logger.debug("check_task_completed: task.uuid: {}".format(check_task.uuid))
            self.watch_task[check_task] = True
            if self.task_lock and self.task_lock.uuid == check_task.uuid:
                self.task_lock = None
            if self.all_tasks_completed(self.watch_task):
                self.becomes_tested()

        check_task.on_task_completed(check_task_completed)
        self.task_queue.insert(0, check_task)
        self.watch_task[check_task] = False

    def _render_task(
        self,
        script_result,  # type: Dict[Text, Any]
        tag,  # type: Text
        visual_grid_manager,  # type: VisualGridRunner
        region_selectors,  # type: List
        region_to_check,  # type: Region
        script_hooks,  # type: Dict[Text, Any]
        check_settings,
    ):
        # type: (...)->RenderTask
        short_description = "{} of {}".format(
            self.configuration.test_name, self.configuration.app_name
        )
        render_task = RenderTask(
            name="RunningTest.render {} - {}".format(short_description, tag),
            script=script_result,
            running_test=self,
            resource_cache=visual_grid_manager.resource_cache,
            put_cache=visual_grid_manager.put_cache,
            rendering_info=visual_grid_manager.render_info(self.eyes),
            eyes_connector=self.eyes,
            region_selectors=region_selectors,
            size_mode=check_settings.values.size_mode,
            region_to_check=region_to_check,
            script_hooks=script_hooks,
            agent_id=self.eyes.base_agent_id,
            selector=check_settings.values.selector,
        )
        logger.debug("RunningTest %s" % render_task.name)

        def render_task_succeeded(render_status):
            # type: (RenderStatusResults) -> None
            logger.debug(
                "render_task_succeeded: task.uuid: {}".format(render_task.uuid)
            )
            if render_status:
                self.eyes.render_status_for_task(render_task.uuid, render_status)
                for vgr in render_status.selector_regions:
                    if vgr.error:
                        logger.error(vgr.error)
                    else:
                        self.regions.append(vgr.to_region())
            self.watch_render[render_task] = True
            if self.all_tasks_completed(self.watch_render):
                self.becomes_rendered()

        def render_task_error(e):
            logger.debug(
                "render_task_error: task.uuid: {}\n{}".format(render_task.uuid, str(e))
            )
            self.pending_exceptions.append(e)
            self.becomes_completed()

        render_task.on_task_succeeded(render_task_succeeded)
        render_task.on_task_error(render_task_error)
        self.render_queue.append(render_task)
        self.watch_render[render_task] = False
        return render_task

    def close(self):
        # type: () -> Optional[Any]
        if self.state == NEW:
            self.becomes_completed()
            return None

        close_task = VGTask(
            "close {}".format(self.browser_info), lambda: self.eyes.close(False)
        )
        logger.debug("RunningTest %s" % close_task.name)

        def close_task_succeeded(test_result):
            logger.debug("close_task_succeeded: task.uuid: {}".format(close_task.uuid))
            self.test_result = test_result
            if callable(self.on_results):
                self.on_results(test_result)

        def close_task_completed():
            # type: () -> None
            logger.debug("close_task_completed: task.uuid: {}".format(close_task.uuid))
            self.watch_close[close_task] = True
            if self.all_tasks_completed(self.watch_close):
                self.becomes_completed()

        def close_task_error(e):
            logger.debug(
                "close_task_error: task.uuid: {}\n{}".format(close_task.uuid, str(e))
            )
            self.pending_exceptions.append(e)

        close_task.on_task_succeeded(close_task_succeeded)
        close_task.on_task_completed(close_task_completed)
        close_task.on_task_error(close_task_error)
        self.close_queue.append(close_task)
        self.watch_close[close_task] = False

    def abort(self):
        self.eyes.abort()

    def all_tasks_completed(self, watch):
        # type: (Dict) -> bool
        if self.state == "completed":
            return True
        return all(state for state in watch.values())



from __future__ import absolute_import

import typing

from selenium.common.exceptions import WebDriverException

from applitools.common import EyesError, Point, logger
from applitools.common.geometry import RectangleSize
from applitools.core import PositionMemento, PositionProvider

from . import eyes_selenium_utils

if typing.TYPE_CHECKING:
    from typing import Optional, Union, Any, Dict
    from applitools.common.utils.custom_types import AnyWebDriver, AnyWebElement
    from . import EyesWebDriver


class SeleniumPositionProvider(PositionProvider):
    """ Encapsulates page/element positioning """

    _JS_GET_ENTIRE_PAGE_SIZE = """
    var width = Math.max(arguments[0].clientWidth, arguments[0].scrollWidth);
    var height = Math.max(arguments[0].clientHeight, arguments[0].scrollHeight);
    return [width, height];
    """
    _JS_GET_CONTENT_ENTIRE_SIZE = """
        var scrollWidth = document.documentElement.scrollWidth;
        var bodyScrollWidth = document.body.scrollWidth;
        var totalWidth = Math.max(scrollWidth, bodyScrollWidth);
        var clientHeight = document.documentElement.clientHeight;
        var bodyClientHeight = document.body.clientHeight;
        var scrollHeight = document.documentElement.scrollHeight;
        var bodyScrollHeight = document.body.scrollHeight;
        var maxDocElementHeight = Math.max(clientHeight, scrollHeight);
        var maxBodyHeight = Math.max(bodyClientHeight, bodyScrollHeight);
        var totalHeight = Math.max(maxDocElementHeight, maxBodyHeight);
        return [totalWidth, totalHeight];";
    """

    def __init__(self, driver, scroll_root_element):
        # type: (EyesWebDriver, AnyWebElement) -> None
        super(SeleniumPositionProvider, self).__init__()
        self._driver = driver
        self._scroll_root_element = eyes_selenium_utils.get_underlying_webelement(
            scroll_root_element
        )
        self._data_attribute_added = False
        logger.debug("Creating {}".format(self.__class__.__name__))

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self._scroll_root_element == other._scroll_root_element
        return False

    def get_entire_size(self):
        # type: () -> RectangleSize
        """
        :return: The entire size of the container which the position is relative to.
        """
        try:
            if self._driver.is_mobile_app:
                width, height = self._driver.execute_script(
                    self._JS_GET_CONTENT_ENTIRE_SIZE
                )
            else:
                width, height = self._driver.execute_script(
                    self._JS_GET_ENTIRE_PAGE_SIZE, self._scroll_root_element
                )
        except WebDriverException as e:
            raise WebDriverException("Failed to extract entire size! \n{}".format(e))
        return RectangleSize(width=width, height=height)

    def _add_data_attribute_to_element(self):
        # type: () -> None
        if not self._data_attribute_added:
            if hasattr(self, "_element"):
                element = self._element
            elif self._scroll_root_element:
                element = self._scroll_root_element
            else:
                element = self._driver.find_element_by_tag_name("html")
            element = eyes_selenium_utils.get_underlying_webelement(element)
            eyes_selenium_utils.add_data_scroll_to_element(self._driver, element)


class ScrollPositionProvider(SeleniumPositionProvider):
    def set_position(self, location):
        # type: (Point) -> Point
        logger.debug(
            "setting position of %s to %s" % (location, self._scroll_root_element)
        )

        scroll_command = (
            "arguments[0].scrollLeft=%d; arguments[0].scrollTop=%d; "
            "return (arguments[0].scrollLeft+';'+arguments["
            "0].scrollTop);" % (location["x"], location["y"])
        )
        self._last_set_position = eyes_selenium_utils.parse_location_string(
            self._driver.execute_script(scroll_command, self._scroll_root_element)
        )
        self._add_data_attribute_to_element()
        return self._last_set_position

    def get_current_position(self):
        # type: () -> Point
        """
        The scroll position of the current frame.
        """
        return eyes_selenium_utils.get_current_position(
            self._driver, self._scroll_root_element
        )


class CSSTranslatePositionProvider(SeleniumPositionProvider):
    def __init__(self, driver, scroll_root_element):
        # type: (EyesWebDriver, AnyWebElement) -> None
        self._last_set_position = Point.ZERO()
        super(CSSTranslatePositionProvider, self).__init__(driver, scroll_root_element)

    def get_current_position(self):
        # type: () -> Optional[Point]
        if self._last_set_position:
            return self._last_set_position

        return eyes_selenium_utils.get_current_position(
            self._driver, self._scroll_root_element
        )

    def set_position(self, location):
        # type: (Point) -> Point
        logger.info(
            "CssTranslatePositionProvider - Setting position to: {}".format(location)
        )
        negated_location = -location

        # fix for CSS stitching in Chrome 78
        self._driver.execute_script(
            (
                "arguments[0].style.transform='translate(10px,-{:d}px';".format(
                    -location.y
                )
            ),
            self._scroll_root_element,
        )

        self._driver.execute_script(
            (
                "arguments[0].style.transform='translate({:d}px,{:d}px';".format(
                    negated_location["x"], negated_location["y"]
                )
            ),
            self._scroll_root_element,
        )
        self._last_set_position = location.clone()
        self._add_data_attribute_to_element()
        return self._last_set_position

    def get_state(self):
        # type: () -> PositionMemento
        state = super(CSSTranslatePositionProvider, self).get_state()
        state.transform = self._driver.execute_script(
            "return arguments[0].style.transform;", self._scroll_root_element
        )
        return state

    def restore_state(self, state):
        # type: (PositionMemento) -> None
        self._driver.execute_script(
            """\
            var originalTransform = arguments[0].style.transform;\
            arguments[0].style.transform = '{}';\
            return originalTransform;""".format(
                state.transform
            ),
            self._scroll_root_element,
        )
        self._last_set_position = None


class ElementPositionProvider(SeleniumPositionProvider):
    def __init__(self, driver, element):
        # type: (AnyWebDriver, Optional[AnyWebElement]) -> None
        super(ElementPositionProvider, self).__init__(driver, element)
        self._element = element

    def get_current_position(self):
        position = Point(
            self._element.get_scroll_left(), self._element.get_scroll_top()
        )
        logger.debug("Current position: {}".format(position))
        return position

    def set_position(self, location):
        # type: (Point) -> Point
        logger.debug("Scrolling element to {}".format(location))
        position = self._element.scroll_to(location)
        logger.debug("Done scrolling element!")
        self._add_data_attribute_to_element()
        return position

    def get_entire_size(self):
        try:
            size = {
                "width": self._element.get_scroll_width(),
                "height": self._element.get_scroll_height(),
            }
        except WebDriverException as e:
            raise WebDriverException("Failed to extract entire size! \n {}".format(e))
        logger.debug("ElementPositionProvider - Entire size: {}".format(size))
        return RectangleSize(**size)



from enum import Enum
from typing import TYPE_CHECKING, Optional

from applitools.common import Point, Region, logger
from applitools.common.utils import image_utils
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.eyes_selenium_utils import (
    get_cur_position_provider,
    get_updated_scroll_position,
)

if TYPE_CHECKING:
    from PIL import Image
    from applitools.selenium.webdriver import EyesWebDriver


class ScreenshotType(Enum):
    VIEWPORT = "VIEWPORT"
    ENTIRE_FRAME = "ENTIRE_FRAME"


def update_screenshot_type(screenshot_type, image, driver):
    # type: ( Optional[ScreenshotType], Image, EyesWebDriver) -> ScreenshotType
    if screenshot_type is None:
        viewport_size = driver.eyes.viewport_size
        scale_viewport = driver.eyes.should_stitch_content

        if scale_viewport:
            pixel_ratio = driver.eyes.device_pixel_ratio
            viewport_size = viewport_size.scale(pixel_ratio)
        if (
            image.width <= viewport_size["width"]
            and image.height <= viewport_size["height"]
        ):
            screenshot_type = ScreenshotType.VIEWPORT
        else:
            screenshot_type = ScreenshotType.ENTIRE_FRAME
    return screenshot_type


def cut_to_viewport_size_if_required(driver, image):
    # type: (EyesWebDriver, Image) -> Image
    # Some browsers return always full page screenshot (IE).
    # So we cut such images to viewport size

    cur_position = driver.eyes._original_scroll_position
    if cur_position:
        top = cur_position.y
    else:
        top = 0

    screenshot_type = update_screenshot_type(None, image, driver)
    if screenshot_type != ScreenshotType.VIEWPORT:
        viewport_size = driver.eyes.viewport_size
        image = image_utils.crop_image(
            image,
            region_to_crop=Region(
                top=top,
                left=0,
                height=viewport_size["height"],
                width=viewport_size["width"],
            ),
        )
    return image


def calc_frame_location_in_screenshot(driver, frame_chain, screenshot_type):
    window_scroll = eyes_selenium_utils.get_default_content_scroll_position(
        frame_chain, driver
    )
    logger.info("Getting first frame...")
    first_frame = frame_chain[0]
    location_in_screenshot = Point.from_(first_frame.location)

    # We only need to consider the scroll of the default content if the screenshot is a
    # viewport screenshot.If this is a full page screenshot, the frame location will not
    # change anyway.
    if screenshot_type == ScreenshotType.VIEWPORT:
        location_in_screenshot = location_in_screenshot.offset(-window_scroll)

    # For inner frames we must calculate the scroll
    inner_frames = frame_chain[1:]
    for frame in inner_frames:
        location_in_screenshot = location_in_screenshot.offset(
            frame.location - frame.parent_scroll_position
        )
    return location_in_screenshot



import typing
import uuid
from typing import Text

import attr

from applitools.core.fluent import GetSelector

if typing.TYPE_CHECKING:
    from applitools.selenium import Eyes, eyes_selenium_utils
    from applitools.common.utils.custom_types import AnyWebElement

__all__ = ("SelectorByElement", "SelectorByLocator")


@attr.s
class SelectorByElement(GetSelector):
    EYES_SELECTOR_TAG = "data-eyes-selector"

    _sel = attr.ib()  # type: AnyWebElement

    def get_selector(self, eyes):
        # type: (Eyes) -> Text
        random_id = uuid.uuid4().hex
        element = eyes_selenium_utils.get_underlying_webelement(self._sel)
        eyes.driver.execute_script(
            "arguments[0].setAttribute('{}', '{}');".format(
                self.EYES_SELECTOR_TAG, random_id
            ),
            element,
        )
        return '[{}="{}"]'.format(self.EYES_SELECTOR_TAG, random_id)


@attr.s
class SelectorByLocator(GetSelector):
    _sel = attr.ib()  # type: Text

    def get_selector(self, eyes):
        # type: (Eyes) -> Text
        element = eyes.driver.find_element(self._sel)
        return SelectorByElement(element).get_selector(eyes)
